from collections import OrderedDict
from datetime import datetime
from typing import Optional

from asset_discovery.exceptions import DuplicateError
from asset_discovery.models import (
    Asset,
    Location,
    AssetPresence,
    AssetPresenceStatus,
    AssetMetadata,
    ReportingStatus,
    AssetIdentification,
    BusType,
    AssetHealthData)
from asset_discovery.repository.json_data_store import JsonDataStore
from asset_discovery.repository.models.json_data import JsonData
from asset_discovery.repository.file_json_provider_factory import FileJsonProviderFactory
from asset_discovery.util.asset_cache_serializer import AssetCacheSerializer
from asset_discovery.util.object_patcher import ObjectPatcher
from asset_discovery.asset_cache.settings_provider import SettingsProvider
from asset_discovery.asset_cache.asset_cache import AssetCache


class AssetCacheManager:
    _MERGED_CACHE_ID = "merged"

    def __init__(self, settings_provider=None, data_provider_factory=None):
        if settings_provider is None:
            settings_provider = SettingsProvider()
        self._settings_provider = settings_provider
        if data_provider_factory is None:
            data_provider_factory = FileJsonProviderFactory()
        self._data_provider_factory = data_provider_factory
        self._implicit_cache = self._create_asset_cache(
            self._MERGED_CACHE_ID,
            self._get_merged_cache_serializer())
        self._merged_cache_timestamp_data_provider = self._create_merged_cache_timestamp_data_provider()
        self._sub_cache_map = OrderedDict()
        self._sub_cache_config_map = OrderedDict()

    @staticmethod
    def _get_merged_cache_serializer():
        # TODO: The inference logic should be cleaned-up (removed)
        # Task: https://ni.visualstudio.com/DevCentral/_workitems/edit/1479109
        return AssetCacheSerializer(
            infer_built_in_bus_type=False,
            infer_present=False,
            infer_empty_temperature_sensors=False
        )

    def _create_asset_cache(self, cache_id, serializer):
        data_provider = self._data_provider_factory.create_data_provider(
            settings_provider=self._settings_provider,
            # The settings provider returns the path to Shared\Skyline\Data\Assets\Cache,
            # but each cache should be stored in its own location under the Cache directory.
            # By passing the extra paths, containing only the <cache_id>, we'll have the following dir:
            # Shared\Skyline\Data\Assets\Cache\<cache_id>
            extra_paths=[cache_id])
        data_store = JsonDataStore(
            json_provider=data_provider,
            serializer=serializer)
        return AssetCache(data_store=data_store)

    def _create_merged_cache_timestamp_data_provider(self):
        return self._data_provider_factory.create_data_provider(
            settings_provider=self._settings_provider,
            extra_paths=[self._MERGED_CACHE_ID])

    def configure_inner_caches(self, inner_cache_configs):
        for cache_config in inner_cache_configs:
            cache_id = cache_config.id
            self._sub_cache_map[cache_id] = self._create_asset_cache(cache_id, AssetCacheSerializer())
            self._sub_cache_config_map[cache_id] = cache_config

    def get_all(self, include_unreported=False):
        assets = self._implicit_cache.list_assets()
        if not include_unreported:
            assets = self._filter_out_unreported_assets(assets)

        self._infer_unspecified_properties(assets)
        return assets

    @staticmethod
    def _filter_out_unreported_assets(assets):
        filtered_assets = []
        for asset in assets:
            if asset.metadata.reporting_status != ReportingStatus.UNREPORTED_RESOURCE:
                filtered_assets.append(asset)
        return filtered_assets

    @staticmethod
    def _infer_unspecified_properties(assets):
        # TODO: The inference logic should be cleaned-up (removed)
        # Task: https://ni.visualstudio.com/DevCentral/_workitems/edit/1479109
        for asset in assets:
            if asset.bus_type is None:
                asset.bus_type = BusType.BUILT_IN_SYSTEM
            if asset.temperature_sensors is None:
                asset.temperature_sensors = []
            if asset.location is None:
                asset.location = Location(state=AssetPresence(asset_presence=AssetPresenceStatus.PRESENT))
            elif asset.location.state is None:
                asset.location.state = AssetPresence(asset_presence=AssetPresenceStatus.PRESENT)

    def get_from_cache(self, cache_id):
        cache = self._sub_cache_map.get(cache_id, None)
        if not cache:
            return []
        return cache.list_assets()

    def reset_cache(self, cache_id, assets, recompute=True):
        """
        This functions resets the cache provided by cache_id,
        by clearing it of all existing elements and storing the provided assets.
        """
        if cache_id == self._MERGED_CACHE_ID:
            cache = self._implicit_cache
        else:
            cache = self._sub_cache_map.get(cache_id, None)
        if not cache:
            return
        cache.empty_cache()
        for asset in assets:
            try:
                cache.add_asset(asset)
            except DuplicateError:
                # We don't support duplicate entries on the Plugin API.
                # Making sure we don't crash if a plugin doesn't follow the rules.
                pass

        if recompute:
            self.recompute_merged_cache()

    def recompute_merged_cache(self):
        existing_assets = {asset.identification: asset for asset in self._implicit_cache.list_assets()}
        merged_assets = {}
        asset_to_sources_map = {}
        for cache_id, cache in self._sub_cache_map.items():
            cache_config = self._sub_cache_config_map[cache_id]
            merge = cache_config.merge
            cache_assets = cache.list_assets()
            for asset in cache_assets:
                identification = asset.identification
                asset = self._merge_asset_if_already_discovered(asset, merged_assets, merge)
                merged_assets[identification] = asset
                data_sources = asset_to_sources_map.get(identification, [])
                data_sources.append(cache_id)
                asset_to_sources_map[identification] = data_sources

        unreported_assets = self._get_unreported_assets(existing_assets, merged_assets)

        assets = list(merged_assets.values())
        assets.extend(unreported_assets)

        timestamp = datetime.utcnow()
        self._recompute_merged_cache_metadata(assets, existing_assets, asset_to_sources_map, timestamp)
        self.reset_cache(self._MERGED_CACHE_ID, assets, recompute=False)
        self._set_updated_timestamp(timestamp)

    def _set_updated_timestamp(self, timestamp):
        timestamp_dict = {"lastUpdatedTimestamp": timestamp.strftime("%m/%d/%Y, %H:%M:%S")}
        uri = self._merged_cache_timestamp_data_provider.get_qualified_uri("merged_cache_last_update_timestamp")
        data = JsonData(uri, timestamp_dict)
        self._merged_cache_timestamp_data_provider.replace(data)

    def _merge_asset_if_already_discovered(self, asset, merged_assets, merge):
        identification = asset.identification
        already_discovered_asset = merged_assets.get(identification, None)
        if already_discovered_asset is not None and merge:
            asset = self._merge_asset_properties(already_discovered_asset, asset)
        return asset

    @staticmethod
    def _merge_asset_properties(already_discovered_asset, asset):
        if asset.location is not None:
            if already_discovered_asset.location is None:
                already_discovered_asset.location = Location()
            already_discovered_asset.location = ObjectPatcher.patch(already_discovered_asset.location, asset.location)
            asset.location = already_discovered_asset.location

        if asset.health is not None:
            if already_discovered_asset.health is None:
                already_discovered_asset.health = AssetHealthData()
            already_discovered_asset.health = ObjectPatcher.patch(already_discovered_asset.health, asset.health)
            asset.health = already_discovered_asset.health

        return ObjectPatcher.patch(already_discovered_asset, asset)

    @staticmethod
    def _get_unreported_assets(existing_assets, reported_assets):
        existing_identifications = set(existing_assets.keys())
        reported_identifications = set(reported_assets.keys())
        unreported_identifications = existing_identifications.difference(reported_identifications)
        return [
            AssetCacheManager._create_not_present_asset(existing_assets[identification])
            for identification
            in unreported_identifications
        ]

    @staticmethod
    def _create_not_present_asset(existing_asset):
        asset = Asset.create_from_identification(existing_asset.identification)
        asset.name = existing_asset.name
        asset.location = Location(state=AssetPresence(asset_presence=AssetPresenceStatus.NOT_PRESENT))
        return asset

    @staticmethod
    def _recompute_merged_cache_metadata(assets, existing_assets, asset_to_sources_map, timestamp):
        for asset in assets:
            identification = asset.identification
            existing_asset = existing_assets.get(identification, None)
            if AssetCacheManager._existing_asset_has_metadata(existing_asset):
                asset_metadata = existing_asset.metadata
            else:
                asset_metadata = AssetMetadata()
            asset_metadata.asset_identification = identification
            asset_metadata.data_sources = asset_to_sources_map.get(identification, [])
            if asset_metadata.last_updated_timestamp is None or asset != existing_asset:
                asset_metadata.last_updated_timestamp = timestamp
            if asset_metadata.version is None:
                asset_metadata.version = 1
            elif asset != existing_asset:
                asset_metadata.version += 1
            asset.metadata = asset_metadata

    @staticmethod
    def _existing_asset_has_metadata(existing_asset):
        return existing_asset is not None and existing_asset.metadata is not None

    def get_asset_metadata(self, identification):
        asset_id_to_metadata_map = self._get_asset_id_to_metadata_map()
        return asset_id_to_metadata_map.get(identification, None)

    def _get_asset_id_to_metadata_map(self):
        return {asset.identification: asset.metadata for asset in self._implicit_cache.list_assets()}

    def get_all_metadata(self):
        all_metadata = self._get_asset_id_to_metadata_map()
        return list(all_metadata.values())

    def get_by_identification(self, asset_identification: AssetIdentification) -> Optional[Asset]:
        asset = self._implicit_cache.get_by_identification(asset_identification)
        if asset is not None:
            self._infer_unspecified_properties([asset])
        return asset
