import logging
import time
from typing import List

from asset_discovery.discovery_engine import DiscoveryEngine
from asset_discovery.models import Asset, AssetIdentification, ReportingStatus

logger = logging.getLogger()


class AddAssetFailedException(Exception):
    pass


class RemoveAssetFailedException(Exception):
    pass


class UpdateAssetFailedException(Exception):
    pass


class DiscoveryEngineClient:
    def __init__(self, discovery_engine=None):
        if not discovery_engine:
            discovery_engine = DiscoveryEngine()
        self._discovery_engine = discovery_engine

    @staticmethod
    def _log_error(e) -> None:
        logger.error(e)

    def register_change_callback(self, callback_id, callback, *args) -> None:
        self._discovery_engine.register_change_callback(callback_id, callback, *args)

    def list_assets(self, **kwargs) -> List[Asset]:
        return self._discovery_engine.list_assets(**kwargs)

    def add_asset(self, asset: Asset, **kwargs) -> None:
        asset_identification = asset.identification
        initial_asset = self._discovery_engine.get_by_identification(asset_identification)
        if initial_asset is not None:
            raise AddAssetFailedException("The asset to add already exists: " + asset_identification.to_str())
        self._discovery_engine.add_asset(asset, **kwargs)
        is_asset_reported = self._wait(lambda: self._discovery_engine.get_by_identification(asset.identification))
        if not is_asset_reported:
            raise AddAssetFailedException("Timed out waiting for the added asset to be reported.")

    def remove_asset(self, asset_identification: AssetIdentification, **kwargs) -> None:
        initial_asset = self._discovery_engine.get_by_identification(asset_identification)
        if initial_asset is None:
            raise RemoveAssetFailedException("The asset to remove cannot be found: " + asset_identification.to_str())
        self._discovery_engine.remove_asset(asset_identification, **kwargs)
        is_asset_unreported = self._wait(lambda: self._is_asset_unreported(asset_identification))
        if not is_asset_unreported:
            raise RemoveAssetFailedException("Timed out waiting for the removed asset to be unreported.")

    def _is_asset_unreported(self, asset_identification: AssetIdentification):
        asset = self._discovery_engine.get_by_identification(asset_identification)
        reporting_status = self._get_reporting_status(asset)
        return reporting_status == ReportingStatus.UNREPORTED_RESOURCE

    @staticmethod
    def _get_reporting_status(asset: Asset):
        if asset and asset.metadata and asset.metadata.reporting_status:
            return asset.metadata.reporting_status
        return ReportingStatus.UNREPORTED_RESOURCE

    def update_asset(self, asset_update, **kwargs) -> None:
        asset_identification = asset_update.identification
        initial_asset = self._discovery_engine.get_by_identification(asset_identification)
        if initial_asset is None:
            raise UpdateAssetFailedException("The asset to update cannot be found: " + asset_identification.to_str())
        version_before_update = self._get_asset_metadata_version(initial_asset)
        self._discovery_engine.update_asset(asset_update, **kwargs)
        is_asset_updated = self._wait(lambda: self._is_asset_updated(asset_identification, version_before_update))
        if not is_asset_updated:
            raise UpdateAssetFailedException("Timed out waiting for the asset to be updated.")

    @staticmethod
    def _get_asset_metadata_version(asset: Asset) -> int:
        if asset and asset.metadata and asset.metadata.version:
            return asset.metadata.version
        return -1

    def _is_asset_updated(self, asset_identification: AssetIdentification, version_before_update: int) -> bool:
        asset = self._discovery_engine.get_by_identification(asset_identification)
        current_version = self._get_asset_metadata_version(asset)
        return current_version > version_before_update

    def self_test(self, asset_identification, **kwargs):
        self._discovery_engine.self_test(asset_identification, **kwargs)

    def self_calibrate(self, asset_identification, **kwargs):
        self._discovery_engine.self_calibrate(asset_identification, **kwargs)

    def reset(self, asset_identification, **kwargs):
        self._discovery_engine.reset(asset_identification, **kwargs)

    def refresh(self, **kwargs):
        self._discovery_engine.refresh(**kwargs)

    def get_metadata(self, **kwargs):
        return self._discovery_engine.get_metadata()

    def _wait(self, evaluator, timeout_seconds=5.0, sleep_time=0.05) -> bool:
        start = time.time()
        if timeout_seconds <= 0:
            return False
        if evaluator():
            return True
        else:
            time.sleep(sleep_time)
            updated_remaining_time = timeout_seconds - (time.time() - start)
            return self._wait(evaluator, updated_remaining_time, sleep_time)
