import asyncio
import logging
from pathlib import Path

import pyinotify
import threading

from asset_discovery.file_monitor.linux_file_changed_event_handler import FileChangedEventHandler
from asset_discovery.file_monitor.start_monitoring_response import StartMonitoringResponse
from asset_discovery.notification.callback_registry import CallbackRegistry

logger = logging.getLogger()


class LinuxFileMonitor:
    def __init__(self, path, file_name=None, file_extension=None, callback_registry=None):
        self._path = path
        self._file_name = file_name
        self._file_extension = file_extension

        self._callback_registry = callback_registry if callback_registry is not None else CallbackRegistry()
        self._notifier = None
        self._watch_manager = None
        self._monitoring_started = False
        self._notification_event_loop = None
        self._monitoring_thread = None

    def __del__(self):
        self.stop_monitoring()

    def stop_monitoring(self):
        if self._notification_event_loop is not None:
            self._notification_event_loop.call_soon_threadsafe(self._notification_event_loop.stop)
            self._notification_event_loop = None
        if self._watch_manager is not None:
            self._watch_manager.close()
            self._watch_manager = None
        if self._notifier is not None:
            self._notifier = None
        if self._monitoring_thread is not None and self._monitoring_thread.is_alive():
            self._monitoring_thread.join(timeout=1.0)
            self._monitoring_thread = None

    def register_change_callback(self, callback_id, change_callback, *args):
        self._callback_registry.register_change_callback(callback_id, change_callback, *args)

    def start_monitoring(self):
        if self._monitoring_started:
            return StartMonitoringResponse.MonitoringAlreadyStarted
        self._ensure_monitored_dir_exists()

        self._monitoring_thread = threading.Thread(target=self._start_monitoring, daemon=True)
        self._monitoring_thread.start()

        return StartMonitoringResponse.Successful

    def _ensure_monitored_dir_exists(self):
        Path(self._path).mkdir(parents=True, exist_ok=True)

    def _start_monitoring(self):
        logger.debug("Discovery Engine - linux_file_monitor.start_monitoring")

        self._watch_manager = pyinotify.WatchManager()
        self._notification_event_loop = asyncio.new_event_loop()
        self._notifier = pyinotify.AsyncioNotifier(
            self._watch_manager,
            self._notification_event_loop,
            default_proc_fun=FileChangedEventHandler(
                handler=self._callback_registry.call,
                file_extension=self._file_extension,
                file_name=self._file_name))

        flags = pyinotify.EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']
        mask = flags['IN_CREATE'] | flags['IN_DELETE'] | flags['IN_DELETE_SELF']
        mask = mask | flags['IN_MODIFY'] | flags['IN_MOVE_SELF']
        self._watch_manager.add_watch(self._path, mask, rec=True, auto_add=True)

        self._notification_event_loop.call_soon(self._set_monitoring_started)
        self._notification_event_loop.run_forever()

    def _set_monitoring_started(self):
        logger.debug('Discovery Engine - linux_file_monitor started')
        self._monitoring_started = True
