import math
from datetime import datetime
from typing import List, Optional
import logging

from asset_discovery.models import (
    Asset,
    AssetHealthData,
    AssetType,
    BusType,
    SelfCalibration,
    ExternalCalibration,
    CalibrationMode,
    CalibrationOperator,
    LifetimePowerOn,
    Location,
    OperationalState,
    AssetPresence,
    AssetPresenceStatus,
    TemperatureSensor)
from asset_discovery.plugins.sysapi_imports.repository.models import (
    AttributeTag,
    AttributeBusType,
    DeviceResetModes,
    IsPresentStatus,
    SelfTestModes,
    ServiceType,
    PropertyBag)
from asset_discovery.plugins.sysapi_imports.repository.fan_sensor_serializer import FanSensorSerializer
from asset_discovery.plugins.sysapi_imports.repository.power_sensor_serializer import PowerSensorSerializer
from asset_discovery.plugins.sysapi_imports.repository.temperature_sensor_serializer import TemperatureSensorSerializer
from asset_discovery.plugins.sysapi_imports.repository.voltage_sensor_serializer import VoltageSensorSerializer
from asset_discovery.plugins.sysapi_imports.util.datetime_utils import DateTimeUtils

logger = logging.getLogger()


class BagSerializer:
    BUS_TYPE_VALUES = {
        AttributeBusType.BuiltInSystem: BusType.BUILT_IN_SYSTEM,
        AttributeBusType.PciPxi: BusType.PCI_PXI,
        AttributeBusType.Usb: BusType.USB,
        AttributeBusType.Gpib: BusType.GPIB,
        AttributeBusType.Vxi: BusType.VXI,
        AttributeBusType.Serial: BusType.SERIAL,
        AttributeBusType.TcpIp: BusType.TCP_IP,
        AttributeBusType.cRio: BusType.CRIO,
        AttributeBusType.Scxi: BusType.SCXI,
        AttributeBusType.cDaq: BusType.CDAQ,
        AttributeBusType.SwitchBlock: BusType.SWITCH_BLOCK,
        AttributeBusType.Scc: BusType.SCC,
        AttributeBusType.FireWire: BusType.FIRE_WIRE,
        AttributeBusType.Accessory: BusType.ACCESSORY,
        AttributeBusType.Can: BusType.CAN,
        AttributeBusType.SwitchBlockDevice: BusType.SWITCH_BLOCK_DEVICE,
        AttributeBusType.Slsc: BusType.SLSC,
    }

    ASSET_PRESENCE_VALUES = {
        IsPresentStatus.Initializing: AssetPresenceStatus.INITIALIZING,
        IsPresentStatus.Unknown: AssetPresenceStatus.UNKNOWN,
        IsPresentStatus.NotPresent: AssetPresenceStatus.NOT_PRESENT,
        IsPresentStatus.Present: AssetPresenceStatus.PRESENT,
    }

    def __init__(self):
        self._temperature_sensor_serializer = TemperatureSensorSerializer()
        self._voltage_sensor_serializer = VoltageSensorSerializer()
        self._power_sensor_serializer = PowerSensorSerializer()
        self._fan_sensor_serializer = FanSensorSerializer()

    def to_bag(self, asset):
        # TODO: add props to bag from asset
        return PropertyBag(asset.location.resource_uri)

    def from_bag(self, bag, possible_parents):
        if not self._is_asset_bag(bag):
            is_network_device = self.is_network_device(bag)

            is_visa = self._is_visa(bag)

            if is_network_device and is_visa:
                resource_uri = bag.resource_uri
                tcp_primary_hostname = bag.get_property_or_default(AttributeTag.TcpPrimaryHostname, None)
                message = self._get_visa_network_device_error_message(resource_uri, tcp_primary_hostname)
                logger.error(message)
            return None
        # noinspection PyBroadException
        try:
            return self._create_asset(bag, possible_parents)
        except Exception:
            # TODO: log this
            pass

    @staticmethod
    def is_network_device(bag: PropertyBag) -> bool:
        connects_to_bus = bag.get_property_or_default(AttributeTag.ConnectsToBusType, None)
        return connects_to_bus == AttributeBusType.TcpIp

    @staticmethod
    def _get_visa_network_device_error_message(resource_uri, tcp_primary_hostname):
        return "Detected problem with visa network device with resource uri {0} and tcp primary hostname {1}.".format(
            resource_uri, tcp_primary_hostname)

    @staticmethod
    def _is_visa(bag):
        visa_driver_name = "ni-visa"
        number_of_drivers = bag.get_property_or_default(AttributeTag.NumberOfDrivers, 0)
        for i in range(number_of_drivers):
            driver = bag.get_property_or_default(AttributeTag.DriverProgName0 + i, "")
            if driver == visa_driver_name:
                return True
        return False

    def _create_asset(self, bag, possible_parents):
        provided_bus_type = self._get_provided_bus_type(bag)
        bus_type_number = bag.get_property_or_default(AttributeTag.ConnectsToBusType, AttributeBusType.BuiltInSystem)
        asset_type = self._get_asset_type(bag)
        temp_sensors = self._create_temperature_sensors(
            bag,
            AttributeTag.SensorTempCount,
            AttributeTag.SensorTempReading0,
            AttributeTag.SensorTempName0)
        ip_addresses = self._get_ip_addresses(bag)
        return Asset(
            model_name=bag.get_property_or_default(AttributeTag.ModelName, ""),
            model_number=self._get_model_number(bag),
            serial_number=bag.get_property_or_default(AttributeTag.SerialNumber, ""),
            vendor_name=bag.get_property_or_default(AttributeTag.VendorName, ""),
            vendor_number=bag.get_property_or_default(AttributeTag.VendorNumber, 0),
            bus_type=self.BUS_TYPE_VALUES[bus_type_number],
            provided_bus_type=provided_bus_type,
            name=bag.get_property_or_default(AttributeTag.DriverUserAlias0, ""),
            asset_type=asset_type,
            firmware_version=bag.get_property_or_default(AttributeTag.FirmwareRevision, ""),
            hardware_version=bag.get_property_or_default(AttributeTag.HardwareRevision, ""),
            visa_resource_name=self._get_visa_resource_name(bag),
            health=self._get_health(bag),
            temperature_sensors=temp_sensors,
            supports_self_calibration=bag.get_property_or_default(AttributeTag.CalIntSupport, None),
            supports_external_calibration=bag.get_property_or_default(AttributeTag.CalExtSupport, None),
            supports_self_test=self._get_supports_self_test(bag),
            supports_reset=self._get_supports_reset(bag),
            self_calibration=self._create_self_calibration(bag),
            is_ni_asset=bag.get_property_or_default(AttributeTag.IsProductNI, False),
            location=self._create_location(bag, possible_parents),
            external_calibration=self._create_external_calibration(bag),
            hostname=bag.get_property_or_default(AttributeTag.TcpPrimaryHostname, None),
            ip_addresses=ip_addresses
        )

    def _get_provided_bus_type(self, bag: PropertyBag) -> Optional[str]:
        provided_bus_type_number = bag.get_property_or_default(AttributeTag.ProvidesBusType, None)
        if provided_bus_type_number is None:
            return None
        return self.BUS_TYPE_VALUES[provided_bus_type_number]

    def _get_asset_type(self, bag: PropertyBag) -> str:
        is_system_controller = self._is_system_controller(bag)
        is_chassis = self._is_chassis(bag)
        if is_system_controller:
            return AssetType.SYSTEM
        if is_chassis:
            return AssetType.CHASSIS
        return AssetType.GENERIC

    @staticmethod
    def _is_asset_bag(bag):
        is_system_controller = BagSerializer._is_system_controller(bag)
        has_serial_number = BagSerializer._has_valid_serial_number(bag)
        is_device = bag.get_property_or_default(AttributeTag.ItemIsDevice, False)
        is_simulated_virtual = bag.get_property_or_default(AttributeTag.SimulatedVirtual, False)

        return (is_system_controller or (has_serial_number and is_device)) and not is_simulated_virtual

    @staticmethod
    def _is_system_controller(bag):
        has_service = bag.get_property_or_default(AttributeTag.ItemHasService, False)
        if not has_service:
            return False
        number_of_services = bag.get_property_or_default(AttributeTag.NumberOfServices, 0)
        for i in range(number_of_services):
            service = bag.get_property_or_default(AttributeTag.ServiceType0 + i, 0)
            if service == ServiceType.LocalSystem:
                return True
        return False

    @staticmethod
    def _has_valid_serial_number(bag):
        serial_number = bag.get_property_or_default(AttributeTag.SerialNumber, "0")
        if not serial_number:
            return False
        if serial_number == "0":
            return False
        return True

    @staticmethod
    def _create_temperature_sensors(bag, count_tag, reading_tag, name_tag, single_reading=False):
        if single_reading:
            sensor_count = 1
        else:
            sensor_count = bag.get_property_or_default(count_tag, 0)
        temp_sensors = []
        for i in range(sensor_count):
            sensor_reading = bag.get_property_or_default(reading_tag + i, float('nan'))
            if math.isnan(sensor_reading):
                continue
            sensor_name = bag.get_property_or_default(name_tag + i, "")
            temp_sensor = TemperatureSensor(name=sensor_name, reading=sensor_reading)
            temp_sensors.append(temp_sensor)
        return temp_sensors

    @staticmethod
    def _get_ip_addresses(bag: PropertyBag) -> Optional[List[str]]:
        ip_count = bag.get_property_or_default(AttributeTag.TcpNumberOfAddr, 0)
        if not ip_count > 0:
            return None
        ip_addresses = []
        for i in range(ip_count):
            ip_address = bag.get_property_or_default(AttributeTag.TcpAddrIP0 + i, None)
            if ip_address is not None:
                ip_addresses.append(ip_address)
        return ip_addresses

    @staticmethod
    def _get_model_number(bag):
        model_number = bag.get_property_or_default(AttributeTag.CSeriesModuleModelCode, 0)
        if model_number == 0:
            model_number = bag.get_property_or_default(AttributeTag.ModelNumber, 0)
        return model_number

    @staticmethod
    def _get_visa_resource_name(bag):
        visa_driver_name = "ni-visa"
        number_of_drivers = bag.get_property_or_default(AttributeTag.NumberOfDrivers, 0)
        for i in range(number_of_drivers):
            driver = bag.get_property_or_default(AttributeTag.DriverProgName0 + i, "")
            if driver == visa_driver_name:
                return bag.get_property_or_default(AttributeTag.DriverItemName0 + i, "")
        return ""

    def _get_health(self, bag):
        now_timestamp = datetime.utcnow()
        cooling_health = self._fan_sensor_serializer.serialize(bag, now_timestamp)
        temp_sensors = self._temperature_sensor_serializer.serialize(bag, now_timestamp)
        voltage_sensors = self._voltage_sensor_serializer.serialize(bag, now_timestamp)
        power_sensors = self._power_sensor_serializer.serialize(bag, now_timestamp)
        power_on_hours = bag.get_property_or_default(AttributeTag.LifetimePowerOnHours, None)
        lifetime_power_on = None
        if power_on_hours is not None:
            lifetime_power_on = LifetimePowerOn(power_on_hours=power_on_hours,
                                                timestamp=now_timestamp)
        state_number = bag.get_property_or_default(AttributeTag.OperationalState, None)
        operational_state = None
        if state_number is not None:
            operational_state = OperationalState(state_number=state_number,
                                                 timestamp=now_timestamp)

        if (not cooling_health and not temp_sensors and not voltage_sensors and not power_sensors and
                not lifetime_power_on and not operational_state):
            return None
        return AssetHealthData(temperature_sensors=temp_sensors,
                               voltage_sensors=voltage_sensors,
                               power_sensors=power_sensors,
                               cooling_health=cooling_health,
                               lifetime_power_on=lifetime_power_on,
                               operational_state=operational_state)

    @staticmethod
    def _get_supports_self_test(bag):
        # A client should treat the absence of this property the same as it having the value None (0).
        self_test_modes = bag.get_property_or_default(AttributeTag.SupportSelfTestModes, SelfTestModes.NoneMode)
        return self_test_modes != SelfTestModes.NoneMode

    @staticmethod
    def _get_supports_reset(bag):
        # A client should treat the absence of this property the same as it having the value None (0).
        reset_modes = bag.get_property_or_default(AttributeTag.SupportDeviceResetModes, DeviceResetModes.NoneMode)
        return reset_modes != DeviceResetModes.NoneMode

    @staticmethod
    def _create_self_calibration(bag):
        supports_self_calibration = bag.get_property_or_default(AttributeTag.CalIntSupport, False)
        if not supports_self_calibration:
            return None
        self_calibration_date = BagSerializer._get_date_for_tag(
            bag,
            AttributeTag.CalIntLastTime,
            DateTimeUtils.minimum_valid_calibration_date)
        if self_calibration_date > DateTimeUtils.minimum_valid_calibration_date:
            temp_sensors = BagSerializer._create_temperature_sensors(
                bag,
                AttributeTag.CalIntTempCount,
                AttributeTag.CalIntLastTempC,
                AttributeTag.CalIntTempName0)
            is_limited = bag.get_property_or_default(AttributeTag.CalIntLastLimited, False)
            return SelfCalibration(
                temperature_sensors=temp_sensors,
                is_limited=is_limited,
                date=self_calibration_date
            )
        return None

    @staticmethod
    def _get_date_for_tag(bag, date_tag, default_date):
        self_calibration_timestamp = bag.get_property_or_default(date_tag, None)
        if self_calibration_timestamp is not None:
            self_calibration_date = DateTimeUtils.convert_sys_api_serialized_date_to_date(self_calibration_timestamp)
        else:
            self_calibration_date = default_date
        return self_calibration_date

    def _create_location(self, bag, possible_parents):
        asset_presence_number = self._get_presence(bag)
        return Location(
            parent=self._get_parent(bag, possible_parents),
            resource_uri=bag.resource_uri,
            slot_number=bag.get_property_or_default(AttributeTag.SlotNumber, -1),
            state=AssetPresence(asset_presence=self.ASSET_PRESENCE_VALUES[asset_presence_number]),
        )

    @staticmethod
    def _get_presence(bag):
        return bag.get_property_or_default(AttributeTag.IsPresentStatus, IsPresentStatus.Present)

    @staticmethod
    def _get_parent(bag, possible_parents):
        connects_to_link_name = bag.get_property_or_default(AttributeTag.ConnectsToLinkName, "")
        return possible_parents.get(connects_to_link_name, "")

    @staticmethod
    def _create_external_calibration(bag):
        supports_ext_calibration = bag.get_property_or_default(AttributeTag.CalExtSupport, False)
        if not supports_ext_calibration:
            return None
        ext_calibration_date = BagSerializer._get_date_for_tag(
            bag,
            AttributeTag.CalExtLastTime,
            DateTimeUtils.minimum_valid_calibration_date)
        if ext_calibration_date > DateTimeUtils.minimum_valid_calibration_date:
            ext_calibration_next_date = BagSerializer._get_date_for_tag(
                bag,
                AttributeTag.CalExtNextTime,
                DateTimeUtils.minimum_valid_calibration_date)
            temp_sensors = BagSerializer._create_temperature_sensors(
                bag=bag,
                count_tag=0,
                reading_tag=AttributeTag.CalExtLastTempC,
                name_tag=0,
                single_reading=True
            )
            return ExternalCalibration(
                temperature_sensors=temp_sensors,
                is_limited=bag.get_property_or_default(AttributeTag.CalExtLastLimited, False),
                date=ext_calibration_date,
                recommended_interval=bag.get_property_or_default(AttributeTag.CalExtRecommendedInterval, -1),
                next_recommended_date=ext_calibration_next_date,
                comments=bag.get_property_or_default(AttributeTag.CalMiscComments, ""),
                entry_type=CalibrationMode.AUTOMATIC,
                operator=CalibrationOperator(display_name="", user_id=""),
                checksum=bag.get_property_or_default(AttributeTag.CalExtChecksum, "")
            )
        return None

    def get_possible_parents_from_bags(self, bags):
        return {
            self._get_provides_link_name(bag): self._get_driver_user_alias(bag)
            for bag in bags
            if bag.contains_property(AttributeTag.ProvidesLinkName)
        }

    @staticmethod
    def _get_provides_link_name(bag):
        return bag.get_property_or_default(AttributeTag.ProvidesLinkName, "")

    @staticmethod
    def _get_driver_user_alias(bag):
        return bag.get_property_or_default(AttributeTag.DriverUserAlias0, "")

    def get_controller_and_chassis(self, bags):
        return [bag for bag in bags if self.is_controller_or_chassis(bag)]

    def is_controller_or_chassis(self, bag):
        return self._is_system_controller(bag) or self._is_chassis(bag)

    def get_pxi_chassis(self, bags):
        return [bag for bag in bags if self._is_pxi_chassis(bag)]

    @staticmethod
    def _is_chassis(bag):
        return bag.get_property_or_default(AttributeTag.ItemIsChassis, False)

    @staticmethod
    def _is_pxi_chassis(bag):
        return BagSerializer._is_chassis(bag) and BagSerializer._is_bus_type_pxi(bag)

    @staticmethod
    def _is_bus_type_pxi(bag):
        return bag.get_property_or_default(AttributeTag.ProvidesBusType, 0) == AttributeBusType.PciPxi

    def is_present(self, bag):
        presence = self._get_presence(bag)
        return presence == IsPresentStatus.Present

    @staticmethod
    def can_delete(bag: PropertyBag) -> bool:
        return bag.get_property_or_default(AttributeTag.SupportItemDelete, False)
