import importlib
import json
import os
import pkgutil

from asset_discovery.util.plugin_manager_base import PluginManagerBase
from asset_discovery.util.plugin_proxy import PluginProxy

PLUGINS_FOUND = True
try:
    import asset_discovery.plugins
except ImportError:
    PLUGINS_FOUND = False


class PluginManager(PluginManagerBase):
    def __init__(self):
        if PLUGINS_FOUND:
            self._load_plugins()
        else:
            self._discovered_plugins = {}

    @staticmethod
    def _iter_namespace(ns_pkg):
        return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")

    def _load_plugins(self):
        plugins = {}
        plugin_class_name = "AssetPlugin"
        modules = self._iter_namespace(asset_discovery.plugins)
        for _, name, _ in modules:
            plugin = importlib.import_module(name)
            if not hasattr(plugin, plugin_class_name):
                continue
            class_ = getattr(plugin, plugin_class_name)
            weight, merge = self._load_config_for_plugin(plugin)
            plugin_id = self._get_plugin_id(plugin)
            instance = class_()
            plugins[plugin_id] = PluginProxy(plugin_id, weight, merge, instance)
        self._discovered_plugins = plugins

    def _load_config_for_plugin(self, plugin):
        config_file_path = self._get_config_file_path_for_plugin(plugin)

        default_weight = 0.0
        default_merge = False
        try:
            with open(config_file_path) as json_file:
                config = json.load(json_file)
                weight = config.get("weight", default_weight)
                merge = config.get("merge", default_merge)
        except (IOError, ValueError):
            weight = default_weight
            merge = default_merge

        return weight, merge

    @staticmethod
    def _get_config_file_path_for_plugin(plugin):
        path_to_module, module_file_name = os.path.split(plugin.__file__)
        file_name, _ = os.path.splitext(module_file_name)
        config_file_name = "{file_name}.config.json".format(file_name=file_name)
        config_file_path = os.path.join(path_to_module, config_file_name)
        return config_file_path

    @staticmethod
    def _get_plugin_id(plugin):
        path_to_module, module_file_name = os.path.split(plugin.__file__)
        file_name, _ = os.path.splitext(module_file_name)
        return file_name

    @property
    def plugins(self):
        return self._discovered_plugins

    def get_sorted_plugins(self, high_to_low=True, min_weight=0.0):
        sorted_plugins = sorted(self._discovered_plugins.items(), key=lambda item: item[1].weight, reverse=high_to_low)
        return {
            plugin_id: plugin
            for plugin_id, plugin in sorted_plugins
            if plugin.weight >= min_weight
        }

    def get_plugin_by_id(self, plugin_id):
        return self.plugins.get(plugin_id)
